//
// Copyright (C) 2001 Simon Greenwold. You are allowed to use or modify the code, but this note must be kept intact.
// 
// AgentType.cpp: implementation of the AgentType class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <maya/MGlobal.h>
#include "MFCPlugin.h"
#include "AgentType.h"
#include "Agent.h"
#include "Constants.h"
#include "Settings.h"
#include "ZoneTypes.h"
#include <iostream>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

AgentType::AgentType()
{

}


AgentType::AgentType(Agent* seedAgent)
{
	
	_name = seedAgent->getLocalType();
	_perZoneType = seedAgent->perZoneType();
	_hasInterface = seedAgent->hasInterface();

	if (_perZoneType) _nZoneTypes = ZoneTypes::zoneTypes->getNumZoneTypes();
	else	_nZoneTypes = 1;

	_deployed.resize(_nZoneTypes);
	_weight.resize(_nZoneTypes);
	_myAgents.resize(_nZoneTypes);
	_impressions.resize(_nZoneTypes);
	
	_defaultDeployment = seedAgent->startingDeployment();
	_defaultWeight = seedAgent->startingWeight();

	for (int i = 0; i < _nZoneTypes; i++) {
		_deployed[i] = _defaultDeployment;
		_weight[i] = _defaultWeight;
		_myAgents[i] = new list<Agent*>;
		_impressions[i] = -1;
	}

	instantiateAgents(seedAgent);
}

void AgentType::instantiateAgents(Agent* seedAgent)
{
	int i, j;
	Agent* newAgent;


	if (_perZoneType) {
		for (j = 0; j < ZoneTypes::zoneTypes->getNumZoneTypes(); j++) {
			for (i = 0; i < _deployed[j]; i++) {
				newAgent = seedAgent->newOne(j);
				newAgent->setAgentType(this);
				_myAgents[j]->push_front(newAgent);
			}
		}
	}
	else {
		for (i = 0; i < _deployed[0]; i++)
			newAgent = seedAgent->newOne(-1);
			newAgent->setAgentType(this);
			_myAgents[0]->push_front(newAgent);
	}
}

void AgentType::changeDeployment(int zoneType, int newDeployed)
{
	int index = zoneType;
	if (index < 0) index = 0;
	if (newDeployed < 0) return;
	Agent* newAgent;

	list<Agent*>* agentList = _myAgents[index];
	if (agentList->size() > newDeployed) {
		while (agentList->size() > newDeployed) {
			delete (*agentList->begin());
			agentList->pop_front();
		}
	}
	else if (agentList->size() < newDeployed) {
		while (agentList->size() < newDeployed) {
			newAgent = (*agentList->begin())->newOne(zoneType);
			newAgent->copyProperties(*agentList->begin());
			newAgent->setAgentType(this);
			agentList->push_front(newAgent);
		}
	}
	_deployed[index] = newDeployed;
}

AgentType::~AgentType()
{
	for (int i = 0; i < _myAgents.size(); i++)
		delete _myAgents[i];
}

int AgentType::getDeployed(int zoneType)
{
	if (zoneType < 0) return _deployed[0];
	if (zoneType > _nZoneTypes) return 0;
	return _deployed[zoneType];
}

int AgentType::getWeight(int zoneType)
{
	if (zoneType < 0) return _weight[0];
	if (zoneType > _nZoneTypes) return 0;
	return _weight[zoneType];
}

double AgentType::getZoneTypeImpression(int zoneType, bool& valid)
{
	list<Agent*>*	tempList;

	double impression;
	double average = 0.0;
	int		numAgents =0;
	int     numTypes = 0;

	if (zoneType < 0)
		zoneType = 0;
	tempList = _myAgents[zoneType];

	for (list<Agent*>::iterator j = tempList->begin(); j != tempList->end(); j++) {
		impression = (*j)->getImpression(valid);
		if (valid) {
			average += impression;
			numAgents++;
		}
	}

	if (numAgents > 0) {
		valid = true;
		average /= numAgents;
	}
	else {
		valid = false;
		average = 0;
	}

	return average;
}

double AgentType::getImpression(int zoneType, bool& valid)
{
/*	for (vector<list<Agent*>*>::iterator j = _myAgents.begin(); j != _myAgents.end(); j++) {
		if((*j)->getZoneType()==zoneType)
			return (*j)->getImpression(valid);
	}
	//valid = false;
	//return -1;
		
			
	if (zoneType < 0) return _impressions[0];
	if (zoneType > _nZoneTypes) return 0;
	//return _impressions[zoneType];*/
	return -1;
}

void AgentType::setDeployed(int deployed, int zoneType)
{
	if (zoneType > _nZoneTypes) return;
	changeDeployment(zoneType, deployed);
}

void AgentType::setWeight(int weight, int zoneType)
{
	if (zoneType < 0) _weight[0] = weight;
	if (zoneType > _nZoneTypes) return;
	_weight[zoneType] = weight;
}

void AgentType::setImpression(double imp, int zoneType)
{
	if (zoneType < 0) _impressions[0] = imp;
	if (zoneType > _nZoneTypes) return;
	_impressions[zoneType] = imp;
}

double AgentType::accumulate(double& totalWeight)
{
	list<Agent*>*	tempList;

	totalWeight = 0.0;
	double result = 0.0;
	bool countType, valid;
	double impression;
	double average;
	int		numAgents;
	int     numTypes = 0;

	for (int i = 0; i < _myAgents.size(); i++) {
		tempList = _myAgents[i];
		countType = false;
		if ((_weight[i] != 0) && (tempList->size() > 0)) {
			average = 0.0;
			numAgents = 0;
			for (list<Agent*>::iterator j = tempList->begin(); j != tempList->end(); j++) {
				impression = (*j)->getImpression(valid);
				if (valid) {
					countType = true;
					average += impression;
					numAgents++;
				}
			}
			if (countType) {
				totalWeight += _weight[i];
				average *= _weight[i];
				average /= numAgents;
				result += average;
				numTypes++;
			}
		}
	}
	if (numTypes > 0) {
		result /= numTypes;
		totalWeight /= numTypes;
	}

	return result;
}


void	AgentType::doInterface(int zoneType)
{
	if (zoneType < 0) zoneType = 0;
	if (zoneType >= _myAgents.size()) return;
	if ((_myAgents[zoneType])->empty()) {
		AfxMessageBox("You must deploy at least one agent in order to change its settings.");
		return;
	}

	list<Agent*>::iterator i = (_myAgents[zoneType])->begin();

	(*i)->doInterface();
	i++;
	for (; i != (_myAgents[zoneType])->end(); i++) {
		(*i)->copyProperties(*(_myAgents[zoneType])->begin());
	}
}

void	AgentType::copyWeightAndDeploymentSettings(int copyZone)
{
	if (copyZone < 0) return;
	for (int i = 0; i < _myAgents.size(); i++) {
		if (i != copyZone) {
			_weight[i] = _weight[copyZone];
			setDeployed(_deployed[copyZone], i);
		}
	}
}

void	AgentType::copyAgentSettingsAcrossAllAgents(Agent* theAgent)
{
	for (int i = 0; i < _myAgents.size(); i++) {
		for (list<Agent*>::iterator j = (_myAgents[i])->begin(); j != (_myAgents[i])->end(); j++) {
			if ((*j) != theAgent)
				(*j)->copyProperties(theAgent);
		}
	}
}


void AgentType::reportSettings()
{
	for (int i = 0; i < _myAgents.size(); i++) {
		if ((_myAgents[i])->size() != 0) {
			(*(_myAgents[i])->begin())->reportSettings();
		}

		Settings::settings->setSetting(_name, "Deployment", _deployed[i], i);
		Settings::settings->setSetting(_name, "Weight", _weight[i], i);
	}
}

void AgentType::receiveSettings()
{
	list<Agent*>::iterator j;
	int tempInt;

	for (int i = 0; i < _myAgents.size(); i++) {
		if ((_myAgents[i])->size() != 0) {
			j = (_myAgents[i])->begin();
			(*j)->receiveSettings();
			j++;
			for (;j != (_myAgents[i])->end(); j++)
				(*j)->copyProperties(*(_myAgents[i])->begin());
		}

		if (Settings::settings->getSetting(_name, "Weight",tempInt, i)) {
			_weight[i] = tempInt;
		}
		else _weight[i] = _defaultWeight;
	
		if (Settings::settings->getSetting(_name, "Deployment",tempInt, i)) {
			if (!_perZoneType) changeDeployment(-1, tempInt);
			else  changeDeployment(i, tempInt);
		}
		else if (!_perZoneType) changeDeployment(-1, _defaultDeployment);
			else  changeDeployment(i, _defaultDeployment);
	}
}

void AgentType::removeZoneType(int which)
{
	if (!_perZoneType) return;

	list<Agent*>* agentList = _myAgents[which];
	while (! agentList->empty()) {
		delete *agentList->begin();
		agentList->pop_front();
	}

	delete agentList;

	for (int i = which + 1; i < _myAgents.size(); i++) {
		_myAgents[i - 1] = _myAgents[i];
		_deployed[i - 1] = _deployed[i];
		_weight[i - 1] = _weight[i];
	}
	_myAgents.resize(_myAgents.size() - 1);
	_deployed.resize(_myAgents.size());
	_weight.resize(_myAgents.size());
	
	for (int j = which; j < _myAgents.size(); j++) {
		agentList = _myAgents[j];
		for (list<Agent*>::iterator k = agentList->begin(); k != agentList->end(); k++) {
			(*k)->setZoneType((*k)->getZoneType() - 1);
		}
	}
}

void AgentType::addZoneType(int after)
{
	if (!_perZoneType) return;

	after++;
	_myAgents.resize(_myAgents.size() + 1);
	_deployed.resize(_myAgents.size());
	_weight.resize(_myAgents.size());

	for (int i = _myAgents.size() - 1; i > after; i--) {
		_myAgents[i] = _myAgents[i - 1];
		_deployed[i] = _deployed[i - 1];
		_weight[i] = _weight[i - 1];
	}

	_myAgents[after] = new list<Agent*>;
	_deployed[after] = _defaultDeployment;
	_weight[after] = _defaultWeight;

	for (i = 0; i < _defaultDeployment; i++) {
		_myAgents[after]->push_back((*_myAgents[0]->begin())->newOne(after));
	}
}

void AgentType::setNumZoneTypes(int newNumTypes, bool refreshSettings)
{
	if (!_perZoneType) return;
	int difference = newNumTypes - _myAgents.size();

	if (difference > 0) {
		for (int i = 0; i < difference; i++) {
			addZoneType(_myAgents.size() - 1);
		}
	}
	else if (difference < 0) {
		for (int i = 0; i < -difference; i++) {
			removeZoneType(_myAgents.size() - 1);
		}
	}

	if (refreshSettings) receiveSettings();
}