//
// Copyright (C) 2001 Simon Greenwold. You are allowed to use or modify the code, but this note must be kept intact.
// 
// Design.h: interface for the Design class.
//
// Defines the individual on which the genetic algorithm operates
// A Design is composed of a genotype, which is the encoded representation
// of the individual which is actually subject to mutations, and a phenotype,
// which is the interpretation of the individual
// The phenotype is constructed from the genotype only when needed
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_DESIGN_H__B350BB9A_E931_11D3_9F59_0000E87A7A07__INCLUDED_)
#define AFX_DESIGN_H__B350BB9A_E931_11D3_9F59_0000E87A7A07__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Phenotype.h"
#include "GeneticUtility.h"
#include <vector>
#include "Representation.h"
#include <list>

class Genotype;
class MayaInfo;
class Analysis;

using namespace std;

class Design  : public Representation
{
public:
	Design(); //Standard constructor - provides minimal initialization
	Design(const std::vector<MayaInfo*>& originalCurves, const std::vector<string>& layerNames); // Constructs with a list of curves
	Design(const Design& oldDesign); //Copy constructor
	virtual ~Design();
	const Design& Design::operator=(const Design& oldDesign);

// Accessors

	Genotype* getGenome() const {return genome;}; //returns a pointer to the genome, which encodes the design
	Phenotype& getPhenome(); //returns a constant reference to the phenome, the interpretted structure encoded by the genome, updating the phenome if the current rendering of the phenome does not match the current genome

	double	getRawFitness();
	double	getStandardizedFitness();
	double	getAdjustedFitness();

	bool	isFitnessUpdated() const {return fitnessUpdated;}; //returns true if the fitness has been synched with the current genome (and phenome)
	bool	isPhenomeUpdated() const;						   //returns true if the phenome has been synched with the current genome

// Mutators

	void	setRawFitness(double newRawFitness);
	void	mutateParam(ParamType typeOfParam);
	void	mutateInsertion(InsertionType typeOfInsertion);
	void	mutateDeletion();
	void	mutateChange(ChangeType typeOfChange);
	void	crossover(int numberOfCrossingPoints, Design* crossoverTemplate);
	void	randomize(int minNumberOfStartingTransformations, int maxNumberOfStartingTransformations);

	Phenotype& interpret(bool startFromZero = false);
	void	erase();
	void	setEraseOnDelete(bool newErase);	// Erase Maya shapes or don't on deletion of deisgn
	void	setFitnessUpdated(bool updated);	// Is fitness up to date?

	void	saveDesign(string& layerName);

	void	reduceNumZoneTypes(int newNumZoneTypes);

private:

	double calculateFitness(bool newErase = false);

// Genotype
	Genotype*	genome;

// Phenotype
	Phenotype*	phenome;

// Fitnesses
	bool	fitnessUpdated;		// Is the fitness up to date?
	double	rawFitness;
	double	standardizedFitness;
	double	adjustedFitness;

	
// Representation Subclass Obligations
public:
	virtual Analysis*	getAnalysis(const string& name);

private:
	virtual void instantiateAnalyses();

	static	bool designAnalysesInstantiated;
	static	list<Analysis*> designAnalyses;

};

#endif // !defined(AFX_DESIGN_H__B350BB9A_E931_11D3_9F59_0000E87A7A07__INCLUDED_)
