//
// Copyright (C) 2001 Simon Greenwold. You are allowed to use or modify the code, but this note must be kept intact.
// 
// Phenotype.cpp: implementation of the Phenotype class.
//
//////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include <maya/MGlobal.h>
#include "Phenotype.h"
#include "Genotype.h"
#include "Shape.h"
#include "MayaUtilities.h"
#include "MELUtilities.h"
#include "ZoneTypes.h"
#include <iostream>
#include <sstream>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Phenotype::Phenotype(Genotype* genome)
{
	_updated = false;
	_genome = genome;
	_eraseOnDelete = true;
	_impressionsValid = false;
	_layerName = getNewLayerName();
	newLayer(_layerName);
}

Phenotype::Phenotype(const Phenotype& oldPhenotype)
{
	_updated = false;
	_genome = oldPhenotype._genome;
	_eraseOnDelete = oldPhenotype._eraseOnDelete;
	_impressionsValid = oldPhenotype._impressionsValid;
	_layerName = oldPhenotype._layerName;
}

Phenotype::~Phenotype()
{
	deleteAllOnLayer(_layerName);
	eval("delete " + _layerName);
}

const Phenotype& Phenotype::operator=(const Phenotype& oldPhenotype) {
	//put copy code here
	return oldPhenotype;
}

void	Phenotype::setGenome(Genotype* genome)
{
	_genome = genome;
	_shapes.clear();
}

Phenotype& Phenotype::interpret(bool startFromZero)
{
	if (!_genome) return *this;
	if (_updated) return *this;

	//cout << "Interpreting " << _layerName << endl;
	if(!layerExists(_layerName))
		newLayer(_layerName);
	setCurrentLayer(_layerName);
	//hideLayer();
		
	
	Shape::_intersectionCheck = 0;

	int numShapes = _genome->getShapes()->size();
	_shapes.clear();
	for (int i = 0; i < numShapes; i++)
		_shapes.push_back((*(_genome->getShapes()))[i]);

	runPrimaryCommands(startFromZero);
	processIntersections(startFromZero);

	for (list<Shape*>::iterator iter = _shapes.begin(); iter != _shapes.end(); iter++) {
		//cout << "About to loft: " << (*iter)->getName() << endl;
		(*iter)->loft(startFromZero);
	}


	//cout << "Intersection checks required: " << Shape::_intersectionCheck << endl;

	_updated = true;
	return *this;
}

void Phenotype::runPrimaryCommands(bool startFromZero)
{
	bool cont = true;

	for (list<Shape*>::iterator iter = _shapes.begin(); iter != _shapes.end(); iter++)
		(*iter)->initializeInterpret();

	while (cont) {
		cont = false;
		for (iter = _shapes.begin(); iter != _shapes.end(); iter++) {
			if ((*iter)->isPrimary())
				if (((PrimaryShape*)(*iter))->runNextCommand()) cont = true;

		}
	}
}

void Phenotype::processIntersections(bool startFromZero)	// THIS FUNCTION KILLS STARTFROMZERO. STARTFROMZERO IS DEAD!!!
{
	list<ShapePair*>::iterator i;

	list<ShapePair*> ipairs, fullyIntersecting;
	findIntersections(_shapes, ipairs);			// Iterate through _shapes, find all intersection pairs
	removeIntersections(ipairs, _shapes);		// Remove from _shapes any shapes that intersect any others

	list<Shape*> nonIntersecting;
	sliceIntersections(ipairs, nonIntersecting, fullyIntersecting);			// Slice all intersections to produce only shapes that intersect nothing and that are fully intersected
	_shapes.splice(_shapes.end(), nonIntersecting);							// Add the non-intersecting shapes back to _shapes
	for (i = ipairs.begin(); i != ipairs.end(); i++)
		delete *i;

	list<Shape*> booleanShapes;
	performBooleans(fullyIntersecting, booleanShapes);
	_shapes.splice(_shapes.end(), booleanShapes);				// Deal with leakage here.
	for (i = fullyIntersecting.begin(); i != fullyIntersecting.end(); i++)
		delete *i;

}

void Phenotype::findIntersections(list<Shape*>& shapeList, list<ShapePair*>& ipairs)
{
	ipairs.clear();


	list<Shape*>::iterator j;
	for (list<Shape*>::iterator i = shapeList.begin(); i != shapeList.end(); i++) {
		j = i;
		j++;
		while(j != shapeList.end()) {
			if ((*i)->intersects(*j))
				ipairs.push_back(new ShapePair(*i, *j));
			j++;
		}
	}
}

void Phenotype::removeIntersections(list<ShapePair*>& ipairs, list<Shape*>& shapeList)
{
	for(list<ShapePair*>::iterator i = ipairs.begin(); i != ipairs.end(); i++) {
		shapeList.remove((*i)->A());
		shapeList.remove((*i)->B());
	}
}

void Phenotype::sliceIntersections(list<ShapePair*>& ipairs, list<Shape*>& nonIntersecting, list<ShapePair*>& fullyIntersecting)
{
	nonIntersecting.clear();
	fullyIntersecting.clear();


	list<list<Shape*>*> allIntersections;
	fillAllIntersections(ipairs, allIntersections);

	list<Slice*> allSlices;

	for (list<list<Shape*>*>::iterator i = allIntersections.begin(); i != allIntersections.end(); i++) {
		julienne(*(*i), allSlices);
	}

	for (i = allIntersections.begin(); i != allIntersections.end(); i++) delete *i;

	list<Slice*>::iterator j, k;
	ShapePair testPair;

	bool intersects, mustTest;
	while(! allSlices.empty()) {

		if (allSlices.size() == 1) {
			nonIntersecting.push_back((*allSlices.begin())->shape);
			delete *allSlices.begin();
			allSlices.pop_front();
		}
		else {
			j = k = allSlices.begin();
			intersects = false;
			mustTest = false;
			for (k++; k != allSlices.end(); k++) {
				testPair.set((*j)->parent, (*k)->parent);
				for (list<ShapePair*>::iterator l = ipairs.begin(); l != ipairs.end(); l++) {
					if (testPair == **l) {
						mustTest = true;
						break;
					}
				}

				if (mustTest) {
						if((*j)->shape->intersects((*k)->shape)) {								// Potential speed up here
						fullyIntersecting.push_back(new ShapePair((*j)->shape, (*k)->shape)); // Only check for intersection between
						delete *j;
						delete *k;
						allSlices.erase(j);									// Slices from parents known to intersect
						allSlices.erase(k);
						intersects = true;
						break;
					}
				}
			}
			if (! intersects) {
				nonIntersecting.push_back((*j)->shape);
				delete *j;
				allSlices.pop_front();
			}
		}
	}
}

void Phenotype::julienne(list<Shape*>& intersections, list<Slice*>& slices)
{

	double start;
	Slice* newSlice;

	list<double> cuts;
	list<Shape*>::iterator i = intersections.begin();

	Shape* original = *i;
	Shape* vegetable = (*i)->fullDuplicate();		// For some reason erasing original makes extrusion impossible
	_eraseList.push_back(vegetable);



	newSlice = new Slice;
	newSlice->parent = original;
	newSlice->shape = vegetable;
	slices.push_back(newSlice);

	if (intersections.size() <= 1) return;

	for (i++; i != intersections.end(); i++) {
		start = (*i)->getZStart();
		cuts.push_back(start);

//		cout << "  Lining up cut at: " << start << endl;

		cuts.push_back(start + (*i)->getZHeight());

//		cout << "  Lining up cut at: " << start + (*i)->getZHeight() << endl;
	}

	cuts.sort();
	cuts.unique();

	Shape* newShape;
	for (list<double>::iterator j = cuts.begin(); j != cuts.end(); j++ ) {
		if (newShape = vegetable->ZCut(*j, UPPER)) {
			newSlice = new Slice;
			newSlice->shape = newShape;
			newSlice->parent = original;
			slices.push_back(newSlice);
			_eraseList.push_back(newShape);
		}
	}
}

void Phenotype::fillAllIntersections(list<ShapePair*>& ipairs, list<list<Shape*>*>& allIntersections)
{
	bool hasList = false;

	for (list<ShapePair*>::iterator i = ipairs.begin(); i != ipairs.end(); i++) {

		hasList = false;
		for(list<list<Shape*>*>::iterator j = allIntersections.begin(); j != allIntersections.end(); j++) {
			if ((*(*j)->begin()) == (*i)->A()) {
				hasList = true;
				break;
			}
		}
		if (!hasList) {
			allIntersections.push_back(new list<Shape*>);
			j = allIntersections.end();
			j--;
			(*j)->push_front((*i)->A());
		}
		(*j)->push_back((*i)->B());

		hasList = false;
		for(j = allIntersections.begin(); j != allIntersections.end(); j++) {
			if ((*(*j)->begin()) == (*i)->B()) {
				hasList = true;
				break;
			}
		}
		if (!hasList) {
			allIntersections.push_back(new list<Shape*>);
			j = allIntersections.end();
			j--;
			(*j)->push_front((*i)->B());
		}
		(*j)->push_back((*i)->A());
	}
}

void Phenotype::performBooleans(list<ShapePair*>& fullyIntersecting, list<Shape*>& booleanShapes)
{

	list<EmergentShape*> newCurves;

	for (list<ShapePair*>::iterator i = fullyIntersecting.begin(); i != fullyIntersecting.end(); i++) {
		newCurves.clear();
		shuffleIntersections((*i)->A()->getShapeMode(), (*i)->A(), (*i)->B(), newCurves);

		for(list<EmergentShape*>::iterator j = newCurves.begin(); j != newCurves.end(); j++) {
			booleanShapes.push_back(*j);
			_eraseList.push_back(*j);
		}
	}
}

void Phenotype::setEraseOnDelete(bool newErase)
{
	_eraseOnDelete = newErase;
	for (list<Shape*>::iterator i = _shapes.begin(); i != _shapes.end(); i++)
		(*i)->setEraseOnDelete(_eraseOnDelete);
}

void	Phenotype::erase()
{	
	for (list<Shape*>::iterator i = _shapes.begin(); i != _shapes.end(); i++) {
		(*i)->erase();
	}

	for (i = _eraseList.begin(); i != _eraseList.end(); i++) {
		(*i)->erase();
		delete *i;
	}
	_eraseList.clear();
	_shapes.clear();
	_updated = false;
}

string	Phenotype::getNewLayerName()
{
	static int layerNr = 0;
	ostringstream layer;
	layer << LAYER_NAME << layerNr++;
	return layer.str();
}

void Phenotype::hideLayer()
{
	setLayerVisibility(_layerName, false);
	//eval("refresh");
}

void Phenotype::initializeImpressionMatrix()
{
	int nZoneTypes = ZoneTypes::zoneTypes->getNumZoneTypes();
	int nAgentTypes = AgentControl::theAgentControl->getTotalAgentTypes();
	_impressions = new vector<vector<double>*>(nAgentTypes);
	for(int i = 0; i < nAgentTypes; i++)	{
		(*_impressions)[i] = new vector<double>(nZoneTypes);
	}
	_impressionsValid = false;
}

void Phenotype::storeImpression(int agentTypeNum, int zoneTypeNum, double impression)
{
	(*(*_impressions)[agentTypeNum])[zoneTypeNum] = impression;
}

bool Phenotype::getImpression(int AgentType, int ZoneType, double& impression)
{
	cout << "Getting impression from " << AgentType << ", " << ZoneType << endl;
	if(!_impressionsValid)
		return false;
	else	{
		impression = (*(*_impressions)[AgentType])[ZoneType];
		return true;
	}
}

void Phenotype::makeImpressionsInvalid()
{
	int nAgentTypes = AgentControl::theAgentControl->getTotalAgentTypes();
	for(int i = 0; i < nAgentTypes; i++)	{
		delete (*_impressions)[i];
	}
	delete _impressions;
	_impressionsValid = false;
}
