//
// Copyright (C) 2001 Simon Greenwold. You are allowed to use or modify the code, but this note must be kept intact.
// 
// Header file created for use with mt19937 random number generator


// Derived from:
/* A C-program for MT19937: Real number version                */
/*   genrand() generates one pseudorandom real number (double) */
/* which is uniformly distributed on [0,1]-interval, for each  */
/* call. sgenrand(seed) set initial values to the working area */
/* of 624 words. Before genrand(), sgenrand(seed) must be      */
/* called once. (seed is any 32-bit integer except for 0).     */
/* Integer generator is obtained by modifying two lines.       */
/*   Coded by Takuji Nishimura, considering the suggestions by */
/* Topher Cooper and Marc Rieffel in July-Aug. 1997.           */

/* This library is free software; you can redistribute it and/or   */
/* modify it under the terms of the GNU Library General Public     */
/* License as published by the Free Software Foundation; either    */
/* version 2 of the License, or (at your option) any later         */
/* version.                                                        */
/* This library is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of  */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.            */
/* See the GNU Library General Public License for more details.    */
/* You should have received a copy of the GNU Library General      */
/* Public License along with this library; if not, write to the    */
/* Free Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA   */ 
/* 02111-1307  USA                                                 */

/* Copyright (C) 1997 Makoto Matsumoto and Takuji Nishimura.       */
/* When you use this, send an email to: matumoto@math.keio.ac.jp   */
/* with an appropriate reference to your work.                     */


#if !defined(RANDOM__INCLUDED_)
#define RANDOM__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <iostream>

using namespace std;

/* Period parameters */  
#define RANDOM_N 624
#define RANDOM_M 397
#define MATRIX_A 0x9908b0df   /* constant vector a */
#define UPPER_MASK 0x80000000 /* most significant w-r bits */
#define LOWER_MASK 0x7fffffff /* least significant r bits */

/* Tempering parameters */   
#define TEMPERING_MASK_B 0x9d2c5680
#define TEMPERING_MASK_C 0xefc60000
#define TEMPERING_SHIFT_U(y)  (y >> 11)
#define TEMPERING_SHIFT_S(y)  (y << 7)
#define TEMPERING_SHIFT_T(y)  (y << 15)
#define TEMPERING_SHIFT_L(y)  (y >> 18)

class Random {
public:

	static void sgenrand(unsigned long seed)
	{
		/* setting initial seeds to mt[RANDOM_N] using         */
		/* the generator Line 25 of Table 1 in          */
		/* [KNUTH 1981, The Art of Computer Programming */
		/*    Vol. 2 (2nd Ed.), pp102]                  */
		mt[0]= seed & 0xffffffff;
		for (mti=1; mti<RANDOM_N; mti++)
			mt[mti] = (69069 * mt[mti-1]) & 0xffffffff;
	};

	static double  genrand()
	{
		unsigned long y;
		static unsigned long mag01[2]={0x0, MATRIX_A};
		/* mag01[x] = x * MATRIX_A  for x=0,1 */

		if (mti >= RANDOM_N) { /* generate N words at one time */
			int kk;

			if (mti == RANDOM_N+1)   /* if sgenrand() has not been called, */
				sgenrand(4357); /* a default initial seed is used   */

			for (kk=0;kk<RANDOM_N-RANDOM_M;kk++) {
				y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
				mt[kk] = mt[kk+RANDOM_M] ^ (y >> 1) ^ mag01[y & 0x1];
			}
			for (;kk<RANDOM_N-1;kk++) {
				y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
				mt[kk] = mt[kk+(RANDOM_M-RANDOM_N)] ^ (y >> 1) ^ mag01[y & 0x1];
			}
			y = (mt[RANDOM_N-1]&UPPER_MASK)|(mt[0]&LOWER_MASK);
			mt[RANDOM_N-1] = mt[RANDOM_M-1] ^ (y >> 1) ^ mag01[y & 0x1];

			mti = 0;
		}
  
		y = mt[mti++];
		y ^= TEMPERING_SHIFT_U(y);
		y ^= TEMPERING_SHIFT_S(y) & TEMPERING_MASK_B;
		y ^= TEMPERING_SHIFT_T(y) & TEMPERING_MASK_C;
		y ^= TEMPERING_SHIFT_L(y);

		//cout << "Random nr: " << (double)y / (unsigned long)0xffffffff << endl;
		return ( (double)y / (unsigned long)0xffffffff ); /* reals */
		/* return y; */ /* for integer generation */
	};


private:

	static unsigned long mt[RANDOM_N]; /* the array for the state vector  */
	static int mti;

};

#endif