//
// Copyright (C) 2001 Simon Greenwold. You are allowed to use or modify the code, but this note must be kept intact.
// 
// Settings.cpp: implementation of the Settings class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <maya/MGlobal.h>
#include "MFCPlugin.h"
#include "Settings.h"
#include "Settable.h"
#include "MELUtilities.h"
#include <iostream>
#include <fstream>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

list<Setting*> Settings::_settings;
string Settings::_settingsFile;
Settings* Settings::settings = new Settings();
bool Settings::zoneTypesInitialized = false;


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Settings::Settings()
{
	setDefaultFile();
	_settings.clear();
	zoneTypesInitialized = false;
}

Settings::~Settings()
{
	clearSettings();
}

string Settings::decodeString(const string& inString)
{
	string result;
	for (int i = 0; i < inString.size(); i++) {
		if (inString[i] != '~') result += inString[i];
		else result += ' ';
	}
	return result;
}

string Settings::encodeString(const string& inString)
{
	string result;
	for (int i = 0; i < inString.size(); i++) {
		if (inString[i] != ' ') result += inString[i];
		else result += '~';
	}
	return result;
}

void Settings::setDefaultFile()
{
	string mayaLoc;
	eval("string $mayaLoc = `getenv \"MAYA_LOCATION\"`");
	getMELvalue("$mayaLoc", mayaLoc);
	for (int i = 0; i < mayaLoc.size(); i++)
		if (mayaLoc[i] == '/') mayaLoc[i] = '\\';

	_settingsFile = mayaLoc + DEFAULT_SETTINGS_FILE;
}

void Settings::readSettings()
{
	ifstream newFile;

	list<Setting*> newSettings;
	newFile.open(_settingsFile.c_str(), ios::in );
	if (newFile.fail()) {
		AfxMessageBox(("Error reading settings file: " + _settingsFile).c_str());
		return;
	}
	
	Setting* newSetting;
	bool fail = false;
	while (!(newFile.eof() || fail)) {
		newSetting = new Setting;

		newFile >> newSetting->type;
		if (newFile.eof()) {
			delete newSetting;
			break;	
		}
		if (newFile.fail()) {
			AfxMessageBox(("Error reading setting type from file: " + _settingsFile).c_str());
			cout << "End of file?" << newFile.eof() << endl;
			
			fail = true;
			delete newSetting;
			break;
		}

		newFile >> newSetting->category;
		if (newFile.fail()) {
			AfxMessageBox(("Error reading setting category from file: " + _settingsFile).c_str());
			fail = true;
			delete newSetting;
			break;
		}

		newFile >> newSetting->name;
		if (newFile.fail()) {
			AfxMessageBox(("Error reading setting name from file: " + _settingsFile).c_str());
			fail = true;
			delete newSetting;
			break;
		}

		if (newSetting->type == "string") {
			newFile >> newSetting->sValue;
			newSetting->sValue = decodeString(newSetting->sValue);
			if (newFile.fail()) {
				AfxMessageBox(("Error reading setting string value from file: " + _settingsFile).c_str());
				fail = true;
				delete newSetting;
				break;
			}
		}
		else if (newSetting->type == "double") {
			newFile >> newSetting->dValue;
			if (newFile.fail()) {
				AfxMessageBox(("Error reading setting double value from file: " + _settingsFile).c_str());
				fail = true;
				delete newSetting;
				break;
			}
		}
		else if (newSetting->type == "int") {
			newFile >> newSetting->iValue;
			if (newFile.fail()) {
				AfxMessageBox(("Error reading setting int value from file: " + _settingsFile).c_str());
				fail = true;
				delete newSetting;
				break;
			}
		}
		else {
			AfxMessageBox(("Error bad value type \"" + newSetting->type + "\" from file: " + _settingsFile).c_str());
			fail = true;
			delete newSetting;
			break;
		}

		newFile >> newSetting->index;
		if (newFile.fail()) {
			AfxMessageBox(("Error reading setting index from file: " + _settingsFile).c_str());
			fail = true;
			delete newSetting;
			break;
		}

		if (!fail) {
			newSettings.push_back(newSetting);
		}
	}
	newFile.close();
	if (fail) {
		while (! newSettings.empty()) {
			delete *newSettings.begin();
			newSettings.pop_front();
		}
	}
	else {
		setSettings(newSettings);
		disseminateSettings();
	}

}

void Settings::writeSettings()
{
	ofstream newFile;

	gatherSettings();
	
	newFile.open(_settingsFile.c_str(), ios::out);
	if (newFile.fail()) {
		AfxMessageBox(("Error writing settings file: " + _settingsFile).c_str());
		return;
	}
	for (list<Setting*>::iterator i = _settings.begin(); i != _settings.end(); i++) {
		newFile << (*i)->type << " " << (*i)->category << " " << (*i)->name << " ";
		if ((*i)->type == "string") newFile << encodeString((*i)->sValue) << " ";
		else if ((*i)->type == "double") newFile << (*i)->dValue << " ";
		else newFile << (*i)->iValue << " ";
		newFile << (*i)->index << "\r\n";
	}
	newFile.close();
}

void Settings::setSettings(list<Setting*>& newSettings)
{
	clearSettings();
	for (list<Setting*>::iterator i = newSettings.begin(); i != newSettings.end(); i++) {
		_settings.push_back(*i);
	}
}

void Settings::clearSettings()
{
	for (list<Setting*>::iterator i = _settings.begin(); i != _settings.end(); i++) {
			delete *i;
	}
	_settings.clear();
}

void Settings::gatherSettings()
{
	list<Settable*>* settables = Settable::getSettables();

	for (list<Settable*>::iterator i = settables->begin(); i != settables->end(); i++) {
		(*i)->reportSettings();
	}
}

void Settings::disseminateSettings()
{	
	list<Settable*>* settables = Settable::getSettables();

	for (list<Settable*>::iterator i = settables->begin(); i != settables->end(); i++) {
		(*i)->receiveSettings();
	}
}

string Settings::getSettingType(const string& category, const string& settingName, int index /*= 0*/)
{
	for (list<Setting*>::iterator i = _settings.begin(); i != _settings.end(); i++) {
		if ((*i)->category == category && (*i)->name == settingName && (*i)->index == index) return (*i)->type;
	}
	return "";
}


string Settings::getStringSetting(const string& category, const string& settingName, int index /*= 0*/)
{
	for (list<Setting*>::iterator i = _settings.begin(); i != _settings.end(); i++) {
		if ((*i)->category == category && (*i)->type == "string" && (*i)->name == settingName && (*i)->index == index) 
			return (*i)->sValue;
	}
	return "";
}

int    Settings::getIntSetting(const string& category, const string& settingName, int index /*= 0*/)
{
	for (list<Setting*>::iterator i = _settings.begin(); i != _settings.end(); i++) {
		if ((*i)->category == category && (*i)->type == "int" && (*i)->name == settingName && (*i)->index == index) 
			return (*i)->iValue;
	}
	return 0;
}

double Settings::getDoubleSetting(const string& category, const string& settingName, int index /*= 0*/)
{
	for (list<Setting*>::iterator i = _settings.begin(); i != _settings.end(); i++) {
		if ((*i)->category == category && (*i)->type == "double" && (*i)->name == settingName && (*i)->index == index) 
			return (*i)->dValue;
	}
	return 0.0;
}

bool Settings::getSetting(const string& category, const string& settingName, string& value, int index /*= 0*/)
{
	for (list<Setting*>::iterator i = _settings.begin(); i != _settings.end(); i++) {
		if ((*i)->category == category && (*i)->type == "string" && (*i)->name == settingName && (*i)->index == index) {
			value = (*i)->sValue;
			return true;
		}
	}
	return false;
}

bool Settings::getSetting(const string& category, const string& settingName, int& value, int index /*= 0*/)
{
	for (list<Setting*>::iterator i = _settings.begin(); i != _settings.end(); i++) {
		if ((*i)->category == category && (*i)->type == "int" && (*i)->name == settingName && (*i)->index == index) {
			value = (*i)->iValue;
			return true;
		}
	}
	return false;
}

bool Settings::getSetting(const string& category, const string& settingName, double& value, int index /*= 0*/)
{
	for (list<Setting*>::iterator i = _settings.begin(); i != _settings.end(); i++) {
		if ((*i)->category == category && (*i)->type == "double" && (*i)->name == settingName && (*i)->index == index) {
			value = (*i)->dValue;
			return true;
		}
	}
	return false;
}

bool Settings::setSetting(const string& category, const string& settingName, const string& setting, int index /*= 0*/)
{
	for (list<Setting*>::iterator i = _settings.begin(); i != _settings.end(); i++) {
		if ((*i)->category == category && (*i)->type == "string" && (*i)->name == settingName && (*i)->index == index) {
			(*i)->sValue = setting;
			return true;
		}
	}
	Setting* newSetting = new Setting;
	newSetting->category = category;
	newSetting->type = "string";
	newSetting->name = settingName;
	newSetting->index = index;
	newSetting->sValue = setting;
	_settings.push_back(newSetting);
	return false;
}

bool Settings::setSetting(const string& category, const string& settingName, int setting, int index /*= 0*/)
{
	for (list<Setting*>::iterator i = _settings.begin(); i != _settings.end(); i++) {
		if ((*i)->category == category && (*i)->type == "int" && (*i)->name == settingName && (*i)->index == index) {
			(*i)->iValue = setting;
			return true;
		}
	}
	Setting* newSetting = new Setting;
	newSetting->category = category;
	newSetting->type = "int";
	newSetting->name = settingName;
	newSetting->index = index;
	newSetting->iValue = setting;
	_settings.push_back(newSetting);
	return false;
}

bool Settings::setSetting(const string& category, const string& settingName, double setting, int index /*= 0*/)
{
	for (list<Setting*>::iterator i = _settings.begin(); i != _settings.end(); i++) {
		if ((*i)->category == category && (*i)->type == "double" && (*i)->name == settingName && (*i)->index == index) {
			(*i)->dValue = setting;
			return true;
		}
	}
	Setting* newSetting = new Setting;
	newSetting->category = category;
	newSetting->type = "double";
	newSetting->name = settingName;
	newSetting->index = index;
	newSetting->dValue = setting;
	_settings.push_back(newSetting);
	return false;
}

void Settings::addSetting(const string& category, const string& settingName, const string& type, int index /*= 0*/)
{
	Setting* newSetting = new Setting;
	newSetting->category = category;
	newSetting->type = type;
	newSetting->name = settingName;
	newSetting->index = index;
	newSetting->iValue = 0;
	newSetting->dValue = 0.0;
	newSetting->sValue = "";
	_settings.push_back(newSetting);
}

void Settings::removeSetting(const string& category, const string& settingName, int index /*= 0*/)
{
	for (list<Setting*>::iterator i = _settings.begin(); i != _settings.end(); i++) {
		if ((*i)->category == category && (*i)->name == settingName && (*i)->index == index) {
			_settings.erase(i);
		}
	}
}

int	 Settings::numberOf(const string& category, const string& settingName)
{
	int count = 0;

	for (list<Setting*>::iterator i = _settings.begin(); i != _settings.end(); i++) {
		if ((*i)->category == category && (*i)->name == settingName)
			count++;
	}
	return count;
}