//
// Copyright (C) 2001 Simon Greenwold. You are allowed to use or modify the code, but this note must be kept intact.
// 
// Shape.cpp: implementation of the Shape class.
//
//////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include <maya/MGlobal.h>
#include "Shape.h"
#include "Constants.h"
#include <maya/MDagPath.h>
#include <maya/MFnTransform.h>
#include <maya/MGlobal.h>
#include <maya/NewMVector.h>
#include <maya/MSelectionList.h>
#include <maya/MItSelectionList.h>
#include <maya/MTransformationMatrix.h>
#include <maya/MFnDagNode.h>
#include <maya/MDagModifier.h>
#include "MELUtilities.h"
#include "MayaUtilities.h"
#include "ZoneTypes.h"
#include <iostream>
#include <sstream>

// Statics

int Shape::_intersectionCheck = 0;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Shape::Shape()
:	ZStart(0.0),
	ZStartScale(1.0),
	ZHeight(STARTING_Z_HEIGHT), 
	ZWidening(STARTING_Z_WIDENING), 
	shapeMode(1),
	hasCut(false),
	cutStart(1),
	cutStop(0)
{
//	cout << "Creating shape: " << this << endl;

	mayaShape = new MayaInfo();
	eraseOnDelete = true;
}

Shape::Shape(MayaInfo* MI)
:	ZStart(0.0),
	ZStartScale(1.0),
	ZHeight(STARTING_Z_HEIGHT), 
	ZWidening(STARTING_Z_WIDENING), 
	shapeMode(1),
	hasCut(false),
	cutStart(1),
	cutStop(0)
{
//	cout << "Creating shape: " << this << endl;
	string curvename = getName();

	mayaShape = MI;
	eraseOnDelete = true;

}

Shape::Shape(const Shape& oldShape)
{
//	cout << "Creating shape: " << this << endl;

	assignment = oldShape.assignment;
	ZWidening = oldShape.ZWidening;
	shapeMode = oldShape.shapeMode;
	ZHeight = oldShape.ZHeight; 
	ZStart = oldShape.ZStart;
	ZStartScale = oldShape.ZStartScale;
	hasCut = oldShape.hasCut;
	cutStart = oldShape.cutStart;
	cutStop = oldShape.cutStop;
	
	mayaShape = new MayaInfo();
	mayaShape->setName(oldShape.mayaShape->name());

	eraseOnDelete = true;
}

Shape::~Shape()
{
//	cout << "Destructing shape: " << this << endl;
	
	if (eraseOnDelete)
		erase(); // first of all, erase any associated Maya objects

	if (mayaShape) delete mayaShape;

}

void Shape::MayaSelect() {
	MStatus				status;

	if (cleanupList.size() == 0) return;

	mayaShape->setName("");

	status = MGlobal::clearSelectionList();
	for (list<string>::const_iterator iter = cleanupList.begin(); iter != cleanupList.end(); iter++) {
		status = MGlobal::selectByName((*iter).c_str());
	}
}

void Shape::erase() {
	MStatus				status;

	if (cleanupList.size() == 0) return;

	if (mayaShape)
		mayaShape->setName("");

	status = MGlobal::clearSelectionList();
	for (list<string>::const_iterator iter = cleanupList.begin(); iter != cleanupList.end(); iter++) {
		status = MGlobal::selectByName((*iter).c_str());
	}
	eval("delete;");

	cleanupList.clear();
}

bool Shape::intersects(Shape* B, bool startFromZero)
{
	_intersectionCheck++;

	if (!Zintersects(B, startFromZero)) return false;
	return  planarIntersect(this, B);
}

bool Shape::Zintersects(Shape* B, bool startFromZero)
{

	double effectiveZStartA, effectiveZStartB;

	if (startFromZero) {
		effectiveZStartA = 0.0;
		effectiveZStartB = 0.0;
	}
	else {
		effectiveZStartA = ZStart;
		effectiveZStartB = B->getZStart();
	}

	// Check Z intersection first (cheap)
	if (effectiveZStartA + ZHeight <= effectiveZStartB) return false;
	if (effectiveZStartB + B->getZHeight() <= effectiveZStartA) return false;

	return true;
}

Shape*	Shape::ZCut(double loc, UpperLower become, bool startFromZero)		// Cuts shape in two at Z = loc, this shape becomes upper or lower depending on value of "become"; the other is returned
{

	double low;

	if (startFromZero) {
		low = 0.0;
		ZHeight += ZStart;
		if (loc <= ZStart) ZStart = 0;
	}
	else {
		low = ZStart;
	}

	double high = low + ZHeight;

	if ((loc <= low) || (loc >= high)) return NULL;

//	cout << endl;
//	cout << "Cutting shape: " << mayaShape->name() << endl;
//	cout << "  Low:  " << low << endl;
//	cout << "  High: " << high << endl;
//	cout << "  At:   " << loc << endl;

	Shape* newShape = fullDuplicate();

	if (become == UPPER) {			// This shape is the top one

		ZWidening = 1.0 - (1.0 - ZWidening) * ((high - loc) / (high - low));  // Proportionalize ZWidening
		ZHeight = high - loc;
		ZStart = loc;

		newShape->ZWidening  = 1.0 - (1.0 - ZWidening) * ((loc - low) / (high - low));
		newShape->ZHeight = loc - low;
		newShape->ZStartScale = 1.0;

		ZStartScale = newShape->ZWidening;

	}
	else  { // Become the lower one

		newShape->ZWidening = 1.0 - (1.0 - ZWidening) * ((high - loc) / (high - low));
		newShape->ZHeight = high - loc;
		newShape->ZStart = loc;

		ZWidening = 1.0 - (1.0 - newShape->ZWidening) * ((loc - low) / (high - low));
		ZHeight = loc - low;
		ZStartScale = 1.0;

		newShape->ZStartScale = ZWidening;
	}

	return newShape;
}

void Shape::setMayaName(string& newName)
{
	mayaShape->setName(newName);
	cleanupList.push_back(newName);
}

void Shape::cut(double start, double stop) {

	if (start == stop) return;
	
	vector<string> newCurves;
	vector<double> cuts(2);
	cuts[0] = start;
	cuts[1] = stop;

	sliceCurve(mayaShape->name(), cuts, newCurves);

	if (newCurves.size() == 2) {
		cleanupList.remove(mayaShape->name());
		eval("delete " + mayaShape->name());

		if (start > stop) {

			mayaShape->setName(newCurves[0]);
			cleanupList.push_back(newCurves[0]);
			eval("delete " + newCurves[1]);
		}
		else {
			mayaShape->setName(newCurves[1]);
			cleanupList.push_back(newCurves[1]);
			eval("delete " + newCurves[0]);
		}
	} 
	else {
		for (int i = 0; i < newCurves.size(); i++) {
			eval("delete " + newCurves[i]);
		}
	}


//	cutCurve(mayaShape->name(), start, stop);
}

bool Shape::loft(bool startFromZero) {
	MStatus				status;
	MDagPath			dup1, dup2;
	MObject				dupObject1, dupObject2, dupTransform1, dupTransform2;
	MVector				translation;
	int					i;

	if (!mayaShape->node()) {
		cout << "No shape named " << mayaShape->name() << endl;
		return false;
	}


	eval("rebuildCurve -kr 0 -rt 3 -ch off " + mayaShape->name());	// Reparameterize curve from 0-1, remove multiple knots
	if (hasCut) cut(cutStart, cutStop);
//	eval("delete -ch " + mayaShape->name());

	MFnDagNode			fnDagNode(*mayaShape->node(), &status);
	if (!status) {
		status.perror("MFnDagNode constructor");
		return false;
	}

	dupObject1 = fnDagNode.duplicate ( FALSE, FALSE, &status );  // I have no idea what these parameters mean
	if (!status) {
		status.perror("MFnDagNode duplicate");
		return false;
	}

	status = MDagPath::getAPathTo ( dupObject1, dup1 ); 
	if (!status) {
		status.perror("MDagPath getAPathTo");
		return false;
	}

	dupTransform1 = dup1.transform(&status);
	// This node has no transform - i.e., it's the world node
	if (!status && status.statusCode () == MStatus::kInvalidParameter)
		return false;

	MFnTransform dupFnTransform1(dupTransform1, &status);
	if (!status) {
		status.perror("MFnTransform constructor");
		return false;
	}

	translation[0] = 0;
	translation[1] = startFromZero ? 0 : ZStart;
	translation[2] = 0;

	status = dupFnTransform1.translateBy(translation, MSpace::kTransform);
	if (!status) {
		status.perror("MFnTransformNode translateBy");
		return false;
	}

	double				scalings[3];
	if (ZStartScale != 1.0) {
		status = dupFnTransform1.getScale(scalings);
		if (!status) {
			status.perror("MFnTransformNode getScale");
			return false;
		}

		scalings[0] *= ZStartScale;
		scalings[2] *= ZStartScale;

		status = dupFnTransform1.setScale(scalings);
		if (!status) {
			status.perror("MFnTransformNode setScale");
			return false;
		}
	}


	dupObject2 = fnDagNode.duplicate ( FALSE, FALSE, &status );  // I have no idea what these parameters mean
	if (!status) {
		status.perror("MFnDagNode duplicate");
		return false;
	}

	status = MDagPath::getAPathTo ( dupObject2, dup2 ); 
	if (!status) {
		status.perror("MDagPath getAPathTo");
		return false;
	}

	dupTransform2 = dup2.transform(&status);
	// This node has no transform - i.e., it's the world node
	if (!status && status.statusCode () == MStatus::kInvalidParameter)
		return false;

	MFnTransform dupFnTransform2(dupTransform2, &status);
	if (!status) {
		status.perror("MFnTransform constructor");
		return false;
	}

	translation[0] = 0;
	translation[1] = ZHeight + (startFromZero ? 0 : ZStart);
	translation[2] = 0;

	status = dupFnTransform2.translateBy(translation, MSpace::kTransform);
	if (!status) {
		status.perror("MFnTransformNode translateBy");
		return false;
	}

	status = dupFnTransform2.getScale(scalings);
	if (!status) {
		status.perror("MFnTransformNode getScale");
		return false;
	}

	scalings[0] *= ZWidening;
	scalings[2] *= ZWidening;

	status = dupFnTransform2.setScale(scalings);
	if (!status) {
		status.perror("MFnTransformNode setScale");
		return false;
	}

	eval("string $shapeList[];");

	ostringstream loftString;
	loftString << "$shapeList = `loft -ch 1 -u 1 -c 0 -ar 0 -d 3 -rn 0 -po 0 -r 1 -r 1 \"" << (dup1.partialPathName (&status)).asChar() << "\" \"" << (dup2.partialPathName (&status)).asChar() << "\"`;";
	eval(loftString.str().c_str());

	ostringstream shadeString;
//	shadeString << "hyperShade -assign shader" << assignment.getType() << ";";

	shadeString << "sets -e -forceElement " << ZoneTypes::zoneTypes->getZoneShader(assignment.getType());
	eval(shadeString.str().c_str());
	eval("displaySmoothness -du 2 -dv 2 -pw 16 -ps 4;");

	vector<string> shapeList;
	getMELvalue("$shapeList", shapeList);
	for (i = 0; i < shapeList.size(); i++) {
		// store lofted surface
		if (shapeList[i].find("loftedSurface") == 0) {
			loftName = shapeList[i];
		}
		cleanupList.push_back(shapeList[i]);
	}
	cleanupList.push_back(dup1.partialPathName(&status).asChar());
	cleanupList.push_back(dup2.partialPathName(&status).asChar());

	return true;
}

void Shape::reduceNumZoneTypes(int newNumZoneTypes)
{
	if (assignment.getType() >= newNumZoneTypes) assignment.mutate();
}